/*
 * Copyright (C) 2013 Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Kaan Onarlioglu -- www.onarlioglu.com
 * NEU SecLab      -- seclab.ccs.neu.edu
 *
 * utils.c, ver.2013.10.13
 *
 * Helper functions for Run-PrivExec.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <pwd.h>
#include <fcntl.h>
#include <ftw.h>
#include <sys/mount.h>

#include "utils.h"

/*
 * Functions for getting/generating directory names.
 */

#define DEFAULT_PRIVEXEC_DIR ".privexec"
#define DEFAULT_STORAGE_DIR "."
#define DEFAULT_TMP_DIR "/tmp"

/*
 * PrivExec works in this subdirectory in the storage and tmp
 * directories below.
 */
const char *get_privexec_dir() {
    const char *dir = getenv( "PRIVEXEC_DIR");
    return dir ? dir : DEFAULT_PRIVEXEC_DIR;
}

/*
 * Overlay mount points are created in this directory. If not user specified in
 * the environment, this searches for the default system tmp directory.
 */
const char *get_tmp_dir() {

    const char *dir = getenv( "PRIVEXEC_TMP");
    if ( !dir) dir = getenv( "TMPDIR");
#if defined P_tmpdir
    if ( !dir) dir = P_tmpdir;
#elif defined _PATH_TMP
    if ( !dir) dir = _PATH_TMP;
#endif
    return dir ? dir : DEFAULT_TMP_DIR;
}

/*
 * Encrypted contents are stored in the storage directory. If not user specified
 * in the environment, this searches for the user's default home directory.
 */
const char *get_storage_dir() {

    const char *dir = getenv( "PRIVEXEC_STORAGE");
    if ( !dir && getenv( "SUDO_UID"))
        return getpwuid( atoi( getenv( "SUDO_UID")))->pw_dir;
    if ( !dir) dir = getenv( "HOME");
    if ( !dir) dir = getpwuid( getuid())->pw_dir;

    return dir ? dir : DEFAULT_STORAGE_DIR;
}

/*
 * This is meant for getting random directory names. Adjust the lookup table
 * accordingly so that you don't get an invalid name for your filesystem/OS.
 */
char *get_random_name( unsigned len) {

    int i;
    static char s[ MAX_RANDOM_LEN + 1]; /* 1 extra for '\0'. */
    static const char *lookup = "1234567890abcdefghijklmnopqrstuvwxyz";
    const size_t lookup_len = strlen( lookup);

    len = ( len < MAX_RANDOM_LEN) ? len : MAX_RANDOM_LEN;

    for ( i = 0; i < len; ++i)
        s[ i] = lookup[ rand() % lookup_len];
    s[i] = '\0';

    return s;
}


/*
 * Exit functions.
 */

/*
 * If this is set, do_exit() will execute it as a final step. Use it as a custom
 * clean up function.
 */
static void ( *cleanup)( void) = NULL;

inline void set_cleanup_func( void ( *fn)( void)) {
    cleanup = fn;
}

/* Prints the libc error and a custom message, and exits. */
void do_exit( int error, char *fmt, ...) {

    va_list args;

    fputs( RED, stdout);
    va_start( args, fmt);
    vprintf( fmt, args);
    va_end( args);
    printf( "Libc says: %s", strerror( error));
    puts( NORMAL);
    fflush( stdout);

    if( cleanup) cleanup(); /* Custom clean up. */

    exit( EXIT_FAILURE);
}


/*
 * String manipulation functions. These always properly allocate/reallocate
 * memory for their return buffers. This makes string manipulation in C slightly
 * more bearable, but if you are running these inside long loops, you are
 * wasting your CPU cycles. Caller frees the string buffers.
 */

/* strdup with error checking. */
void do_strdup( char **dest, const char *src) {
    *dest = strdup( src);
    if ( !*dest) die( "Cannot strdup.\n");
}

/* strcat with error checking, and without the hassle. */
void do_strcat( char **dest, const char *src) {

    char *ret = realloc( *dest, strlen( *dest) + strlen( src) + sizeof( *src));
    if ( !ret) die( "Cannot strcat.\n");

    *dest = ret;
    strcat( *dest, src); /* No need for strncat. */
}

/*
 * Trims spaces at both ends of a string, in place. Returns the input pointer
 * back to you.
 */
char *trim( char *s) {

    char *i = s;
    char *j = s;

    /* Trim left. */
    for( ; *j == ' '; ++j);
    while ( *j != '\0') {
        *i++ = *j++;
    }
    *i = '\0';

    /* Trim right. */
    i = s;
    for( ; *i != '\0'; ++i);
    for( --i; *i == ' ' && i != (s - 1); --i);
    *( i + 1) = '\0';

    return s;
}


/*
 * File and directory manipulation functions.
 */

/*
 * Reads a text file, returns contents in a newly allocated buffer, and buffer
 * size in out argument buf_len. Caller frees the buffer.
 */
char *read_text_file( char *path, unsigned *buf_len) {

    int f;
    char *buf;
    unsigned count;
    unsigned cur;
    unsigned max;

    f = open( path, O_RDONLY);
    if ( f == -1)
        die( "Cannot open file for reading, %s\n", path);

    cur = 0;
    max = IO_CHUNK;
    buf = malloc( max * sizeof(*buf));
    while ( ( count = read( f, buf + cur, IO_CHUNK))){
        if ( count == -1)
            die( "Cannot read from file %s\n", path);

        cur += count;
        buf = try_realloc( buf, cur, &max, IO_CHUNK, sizeof( *buf));
    }
    buf[ cur] = '\0';
    close( f);

    if ( buf_len)
        *buf_len = cur;
    return buf;
}


/*
 * These two walk the specified directory tree and remove all normal files and
 * directories found. _delete() is for internal use by do_remove_dir_tree().
 */
static int _delete( const char *path, const struct stat *sb,
                    int type, struct FTW *buf) {
    switch ( type) {
    case FTW_F:
        unlink( path);
        break;
    case FTW_DP:
        rmdir( path);
        break;
    }
    return 0;
}

void do_remove_dir_tree( char *path) {

    int open_max = sysconf( _SC_OPEN_MAX);
    if ( open_max == -1) open_max = 256;

    if ( nftw( path, _delete, open_max, FTW_DEPTH | FTW_PHYS))
        print_error( "Warning: Dir tree %s removal encountered error(s).\n", path);
}

/* Remove the given list of empty directories. */
void do_remove_dirs( char **dirs) {
    int i;
    for ( i = 0; dirs[ i]; ++i)
        rmdir( dirs[ i]);
}


/*
 * Simple mount and unmount wrappers.
 */
inline void do_umount( char *mnt) {
#ifdef PRIVEXEC_DEBUG
    print_green( "UMNT: %s\n", mnt);
#endif
    umount2( mnt, MNT_DETACH);
}

/* For an explanation of arguments, "man 2 mount". */
inline void do_mount( char *src, char *target,
                      char *fs, unsigned flags, char *opt) {
#ifdef PRIVEXEC_DEBUG
    print_green( "MNT: %s --> %s\n", src, target);
#endif
    if ( mount( src, target, fs, flags, opt))
        die( "Cannot mount %s --> %s\n", src, target);
}


/*
 * Memory management functions.
 */

/*
 * Checks a buffer's size, then allocates more memory (i.e. size * inc bytes),
 * if necessary.
 *
 * buf: buffer to check
 * cur: currently used size of the buffer
 * (in/out) max: maximum cpacity of the buffer, updated to reflect the new size
 *          after a reallocation
 * inc: buffer size is increased to accomodate this many more elements size:
 * size of an element in the buffer, in bytes
 *
 * Return: the pointer to the new larger buffer
 */
void *try_realloc( void *buf, unsigned cur, unsigned *max,
                   unsigned inc, unsigned size) {
    void *new_buf;
    if ( cur == *max) {
        new_buf = realloc( buf, ( *max + inc) * size);
        if( !new_buf) die( "Realloc failed.\n");
        *max += inc;
        return new_buf;
    }
    return buf;
}

/* Frees if not NULL. */
inline void do_free( void *mem) {
    if ( mem) free( mem);
}

/* Frees an array of strings and all strings in it. */
void do_free_double( char **arr, unsigned len) {
    int i;
    if ( len) {
        for ( i = 0; i < len; ++i) {
            free( arr[ i]);
        }
        free( arr);
    }
}
